UnitTest {
	run {
		var methods = this.class.methods;
		
		("Running tests for"+this.class.name).postln;
			
		methods = all {:x, x <- methods, x.name.asString.beginsWith("test")};
		methods.do({|x|
			var failed = false;

			this.tryPerform(\setUp);
			
			try { this.perform(x.name);} { |exception|
				if(exception.isKindOf(AssertionFailure)) {
					("   FAILED:" + x.name.asString).postln;
					("      assertion:" + exception.what.asString ++ ";" + exception.msg).postln;
				} {
					exception.reportError;	
				};
				failed = true;
			};
			
			if(failed == false) {
				("   PASSED:" + x.name.asString).postln;
			};
			
			this.tryPerform(\tearDown);
		});
		
		"".postln;
	}
	
	*runAll {
		var subclasses = UnitTest.allSubclasses;
		
		subclasses.do({ |x|
			var test = x.new;
			test.run;
		});
	}
	
	assertNotNil { |a1, str|
		if(a1 == nil) {
			AssertionFailure(str).throw;
		};
	}

	assertTrue { |a1, str|
		if(a1 != true) {
			AssertionFailure(str).throw;
		};
	}

	assertFalse { |a1, str|
		if(a1 != false) {
			AssertionFailure(str).throw;
		};
	}

	assertIdenticalObjects { |a1, a2, str|
		if(a1 !== a2) {
			AssertionFailure(str).throw;
		};
	}

	assertEquals { |a1, a2, str|
		if(a1 != a2) {
			AssertionFailure(str).throw;
		};
	}

	assertThrows { |a1, str|
		try {a1.value} { ^nil };
		AssertionFailure(str).throw;
	}

	assertThrowsSpecific { |a1, a2, str|
		try {a1.value} { |exception|
			if(exception.class == a2) { ^nil };
		};
		AssertionFailure(str).throw;
	}
	
	assertNoThrows { |a1, str|
		try {a1.value} {
			AssertionFailure(str).throw;
		};
	}

	assertNoThrowsSpecific { |a1, a2, str|
		try {a1.value} { |exception|
			if(exception.class == a2) {AssertionFailure(str).throw};
		};
	}

	fail { |str|
		AssertionFailure(str).throw;
	}

	testAssertNotNil {
		this.assertNoThrows({this.assertNotNil(true, nil)}, "assertNotNil raised on true");
		this.assertThrowsSpecific({this.assertNotNil(nil, nil)}, AssertionFailure, "assertNotNil did not raise on nil");
	}

	testAssertTrue {
		this.assertNoThrows({this.assertTrue(true, nil)}, "assertTrue raised on true");
		this.assertThrowsSpecific({this.assertTrue(false, nil)}, AssertionFailure, "assertTrue did not raise on false");
	}

	testAssertFalse {
		this.assertNoThrows({this.assertFalse(false, nil)}, "assertFalse raised on false");
		this.assertThrowsSpecific({this.assertFalse(true, nil)}, AssertionFailure, "assertFalse did not raise on true");
	}
	
	testAssertIdenticalObjects {
		var obja = "teststr", objb = obja, objc = "teststr";
		this.assertNoThrows({this.assertIdenticalObjects(obja, objb, nil)}, "assertIdenticalObjects raised on identical objects");
		this.assertThrowsSpecific({this.assertIdenticalObjects(obja, objc, nil)}, AssertionFailure, "assertIdenticalObjects did not raise on unidentical objects");
	}
	
	testAssertEquals {
		var obja = "teststr", objb = "teststr", objc = "teststr2";
		this.assertNoThrows({this.assertEquals(obja, objb, nil)}, "assertEquals raised on equal objects");
		this.assertNoThrows({this.assertEquals(nil, nil, nil)}, "assertEquals raised on equal (nil) objects");
		this.assertThrowsSpecific({this.assertEquals(obja, objc, nil)}, AssertionFailure, "assertEquals did not raise on unequal objects");
		this.assertThrowsSpecific({this.assertEquals(obja, nil, nil)}, AssertionFailure, "assertEquals did not raise on unequal objects (where one was nil)");
	}
	
	testFail {
		this.assertThrowsSpecific({this.fail(nil)}, AssertionFailure, "fail did not raise an exception");
	}
}